//
// Created by robin on 1/8/18.
//

#include "dns-base/QDnsPacket.h"

namespace QDns
{
    namespace Base
    {
        QDnsPacket::QDnsPacket()
                : m_isValid(true)
        {
        }

        QDnsPacket::QDnsPacket(const QByteArray &data)
                : QDnsPacket()
        {
            QDataStream dataStream(data);
            parse(data, dataStream);
        }

        QDnsPacket::QDnsPacket(const QByteArray &data, QDataStream &dataStream)
                : QDnsPacket()
        {
            parse(data, dataStream);
        }

        void QDnsPacket::parse(const QByteArray &data, QDataStream &dataStream)
        {
            m_header = QDnsHeader(dataStream);
            if (!m_header.isValid())
            {
                m_isValid = false;
            }
            else
            {
                for (quint16 i = 0; i < m_header.getQueryCount() && m_isValid; ++i)
                {
                    auto entry = QDnsQuery(data, dataStream);
                    if (!entry.isValid())
                    {
                        m_isValid = false;
                    }
                    m_queries.append(entry);
                }
                for (quint16 i = 0; i < m_header.getAnswerCount() && m_isValid; ++i)
                {
                    auto entry = QDnsResourceRecord(data, dataStream);
                    if (!entry.isValid())
                    {
                        m_isValid = false;
                    }
                    m_answerRecords.append(entry);
                }
                for (quint16 i = 0; i < m_header.getAuthorityCount() && m_isValid; ++i)
                {
                    auto entry = QDnsResourceRecord(data, dataStream);
                    if (!entry.isValid())
                    {
                        m_isValid = false;
                    }
                    m_authorityRecords.append(entry);
                }
                for (quint16 i = 0; i < m_header.getAdditionalCount() && m_isValid; ++i)
                {
                    auto entry = QDnsResourceRecord(data, dataStream);
                    if (!entry.isValid())
                    {
                        m_isValid = false;
                    }
                    m_additionalRecords.append(entry);
                }
            }
        }

        QByteArray QDnsPacket::serialize() const
        {
            QByteArray byteArray;
            QDataStream dataStream(&byteArray, QIODevice::WriteOnly);
            serialize(dataStream);
            return byteArray;
        }

        void QDnsPacket::serialize(QDataStream &dataStream) const
        {
            m_header.serialize(dataStream);
            for (const auto &entry : m_queries)
            {
                entry.serialize(dataStream);
            }
            for (const auto &entry : m_answerRecords)
            {
                entry.serialize(dataStream);
            }
            for (const auto &entry : m_authorityRecords)
            {
                entry.serialize(dataStream);
            }
            for (const auto &entry : m_additionalRecords)
            {
                entry.serialize(dataStream);
            }
        }

        bool QDnsPacket::isValid() const
        {
            return m_isValid;
        }

        QDnsPacket QDnsPacket::makeAnswer() const
        {
            auto packet = *this;
            auto header = packet.getHeader();
            header.setIsQuery(false);
            header.setIsRecursionAvailable(false);
            header.setIsAnswerAuthenticated(false);
            packet.setHeader(header);
            packet.m_answerRecords.clear();
            packet.m_authorityRecords.clear();
            packet.m_additionalRecords.clear();
            return packet;
        }

        const QDnsHeader &QDnsPacket::getHeader() const
        {
            return m_header;
        }

        void QDnsPacket::setHeader(const QDnsHeader &header)
        {
            m_header = header;
            m_header.setQueryCount((quint16) m_queries.length());
            m_header.setAnswerCount((quint16) m_answerRecords.length());
            m_header.setAuthorityCount((quint16) m_authorityRecords.length());
            m_header.setAdditionalCount((quint16) m_additionalRecords.length());
        }

        const QList<QDnsQuery> &QDnsPacket::getQueries() const
        {
            return m_queries;
        }

        void QDnsPacket::setQueries(const QList<QDnsQuery> &queries)
        {
            m_queries = queries;
            m_header.setQueryCount((quint16) m_queries.length());
        }

        const QList<QDnsResourceRecord> &QDnsPacket::getAnswerRecords() const
        {
            return m_answerRecords;
        }

        void QDnsPacket::setAnswerRecords(const QList<QDnsResourceRecord> &answerRecords)
        {
            m_answerRecords = answerRecords;
            m_header.setAnswerCount((quint16) m_answerRecords.length());
        }

        const QList<QDnsResourceRecord> &QDnsPacket::getAuthorityRecords() const
        {
            return m_authorityRecords;
        }

        void QDnsPacket::setAuthorityRecords(const QList<QDnsResourceRecord> &authorityRecords)
        {
            m_authorityRecords = authorityRecords;
            m_header.setAuthorityCount((quint16) m_authorityRecords.length());
        }

        const QList<QDnsResourceRecord> &QDnsPacket::getAdditionalRecords() const
        {
            return m_additionalRecords;
        }

        void QDnsPacket::setAdditionalRecords(const QList<QDnsResourceRecord> &additionalRecords)
        {
            m_additionalRecords = additionalRecords;
            m_header.setAdditionalCount((quint16) m_additionalRecords.length());
        }

        QDnsPacket QDnsPacket::makeSimpleQuery(const QString& host, RecordType type)
        {
            QDns::Base::QDnsPacket packet;
            auto header = packet.getHeader();
            header.setIsQuery(true);
            header.setIsRecursionDesired(true);
            packet.setHeader(header);
            QDns::Base::QDnsQuery query;
            query.setName(host);
            query.setClass(QDns::Base::RecordClass::IN);
            query.setType(type);
            packet.setQueries(QList<QDns::Base::QDnsQuery>() << query);
            return packet;
        }
    }
}